__author__ = "Joan Prat"
__copyright__ = "Copyright 2019, The Dimensigon project"
__credits__ = ["Joan Prat", "Daniel Moya"]
__license__ = "GNU General Public License v3 or later (GPLv3+)"
__version__ = "0.3.5"
__maintainer__ = "Joan Prat"
__email__ = "joan.prat@dimensigon.com"
__status__ = "development"
